"""
Lyrics providers for spotdl.
"""

from spotdl.providers.lyrics.genius import Genius
from spotdl.providers.lyrics.musixmatch import MusixMatch
from spotdl.providers.lyrics.azlyrics import AzLyrics
